unit simplwin;
{
SIMPLWIN - a simple text mode windowing support unit.

CHARACTERISTICS:

-SimplWin provides for single statement creation & disposition of text
windows, with borders, shadows, & user selected text attribute in a unit
of less than 5K.
-Original, underlying screen information is saved & restored, as are cursor
positions & screen characteristics. (Overlaid screen data is stored on the
heap)
-Windows are constrained to the current screen limits.  Windows larger than
available screen space are reduced to fit (this is not the same as clipping)
-Window screen settings stay in effect until the window is disposed of, or
until the OUTER method is invoked.  Restoration of a window's boundary can
be accomplished via the INNER method.
-OUTER & INNER also track & restore last cursor position within a window.
-Shadows are clipped at screen boundaries.
-Shadows are oriented to match TV TWindow shadows. (useful in external progs
called by a TV application without a screen clear)

A few other procedures (screen text color & position manipulation) were
tossed in as extras.

SimplWin works with BP 7 real and protected modes.

SimplWin is an objectified distillation of windowing routines I first put
together in '85 as components for a pilot information system user interface I
was developing for the government entity that provides my primary employment.
(By the way, that system has been distributed via CD-ROM since 6/90, and now
uses an interface based on the TurboVision application framework.)

(C) March 1994,
Schoeneck Howell (aka SK Howell/Nick/Nick-the-fish)

showell@bop.gov
skhowell@earthlink.net

Consider it freeware, with all rights retained.
}

interface

uses crt;

type
  windoptr = ^windo;
  windo = object
    topc,topr,
    botc,botr  : integer;
    vbuff      : pointer;
    vbuffsize  : word;
    oldtattr   : byte;
    winx,winy  : byte;
    oldx,oldy  : byte;
    oldWmin    : word;
    oldWmax    : word;
    wtxtattr   : byte;
    wmin       : word;
    wmax       : word;
    bcolor     : byte;
    tcolor     : byte;
    wstyle     : byte;
    constructor init(tc,tr,bc,br,style,tclr,bclr : byte);
      { ===============
        tc - left column (X)
        tr - top row (Y)
        bc - right column
        br - bottom row
        (i.e top left corner/bottom right corner definition)

        style - 0 no border characters - full window available
                1 double line border
                2 single line border
                3 single line side, double line top & bottom
                4 double line side, single line top & bottom
                5 double line top, single line sides & bottom
                6 typewriter border, * in corners, - hor., | vert
                7 typewriter border, + in corners, - hor., | vert
                8 typewriter border, * in corners, = hor., I vert
                9 heavy block border
                styles 0-9 contain shadows
                styles 10-19 are the same styles without shadows
        tclr - textcolor
        bclr - background color
        ==================      }
    destructor done;
    procedure SHADOW(col,row,span : byte);
    procedure wind_brdr;
    procedure outer;
    procedure inner;
  end;

{a window object ... contains the text obscureed in a buffer }
{& remembers pre-window screen & cursor conditions ... which }
{all get restored upon window destruction                    }

procedure cwrite(hue : byte; txt : string);
{ set color & write   }

procedure cwriteln(hue : byte; txt : string);
{ set color & writeln }

procedure vidreverse(c,r,span : integer);
{ reverse the foreground / background colors for a portion of a line }

procedure color_char(col,row,color,span : byte);
{ set the foreground color for a portion of a line}

procedure put_text(col,row : integer; txt : string);
{ combo gotoxy & write}

procedure put_line(col,row : integer; txt : string);
{ combo gotoxy & writeln}

procedure wait_for_a_key(var keyread : char);
{ wait & return a keypress}

procedure wait_for_any_key;
{ wait for a keypress ... then discard it}

implementation

var

  vidseg : word;

{=============================== CHR_IN =============================}
{  j is the offset into video mem. ... vchr will hold the result}
procedure chr_in(j : integer; var vchr : char);
begin;
vchr := chr(mem[vidseg : j]);
end;

{=============================== CHR_OUT =============================}
{  j is the offset into video mem. ... vasc is the ascii code to be output}
procedure chr_out(j,vasc : integer);
begin;
mem[vidseg : j] := vasc;
end;

{=============================== WRD_IN =============================}
{  j is the video mem. offset, vwrd is the 2 byte char(ascii)/attrib holder }
procedure wrd_in(j : integer;var vwrd : word);
begin;
vwrd := memw[vidseg : j];
end;

{=============================== WRD_OUT =============================}
{  j is the video mem. offset, vwrd is the 2 byte char(ascii)/attrib holder }
procedure wrd_out(j : integer; vwrd : word);
begin;
memw[vidseg : j] := vwrd;
end;

{=============================== VID_OFFSET =============================}
{ rowint is row #, colint is column #, func will return the calculated video offset}
function vid_offset(colint,rowint : integer) : integer;
var j: integer;
begin;
j := ((rowint - 1) * 160) + ((colint - 1) * 2);
vid_offset := j;
end;

{=============================== VMEM_OFFSET =============================}
{ rowint is row #, colint is column #, j will hold the calculated video offset}
procedure vmem_offset(colint,rowint : word;var j : word);
begin;
j := ((rowint-1) * 160) + ((colint-1) * 2);
end;

{=============================== BRDR_LINE =============================}
{c & r are col and row, n is length, vchr is line character,
 and d indicates whether left to right (1) or top to bottom (0)}
procedure brdr_line(c,r,n,vchr,d : integer);
var
  i,j : word;
begin;
  vmem_offset(c,r,j);
  for i := 1 to n do
     begin
        chr_out(j,vchr);
        if d = 0 then
           j := j + 160
        else
           j := j + 2;
      end;
end;

{=============================  VIDREVERSE ===============================}
procedure vidreverse(c,r,span : integer);
var
   i,j,k  : word;
   f,b,h  : byte;
begin
   vmem_offset(c,r,j);
   wrd_in(j,k);
   f := hi(k) and 15; {the high byte is video attributes ... half is
                       foreground (the lower 4 bits) and half              }
   b := hi(k) and 240;{is background ( ...the upper 4 bits)                }
   h := 8 and f;      {is the foreground hi intensity? h = 8 if yes
                       0 if no                                             }
   f := f shl 4 ;     {the and's above removed set bits (1's) ... allowing }
   b := b shr 4 ;     {shifts to occur without range errors                }
   b := b and 7;      {strip the intensity bit                             }
   b := b + h;        {restore what was there previously                   }
   f := f and 127;    {make sure there will be no blinking                 }
   for i := 1 to span do chr_out(j+(i*2)-1,f+b);
   {the result of these shifts is to swap foreground/background nibbles}
end;
{==================== COLOR_CHAR ==================}
procedure color_char(col,row,color,span : byte);
var
  k : word;
  i,j : word;
  b,f : byte;
begin
  vmem_offset(col,row,i);
  wrd_in(i,k);
  f:= color;
  b:= hi(k) and 240;      {upper half is background ... 4 bits }
  for j := 1 to span do
   chr_out(i+(2*j)-1,f+b); {plug mod back into memory .....}
end;

{ ===================== cwrite ================================== }
procedure cwrite(hue : byte; txt : string);
 {set color & write   }
var oldcolor : byte;
begin
  oldcolor:=textattr;
  textcolor(hue);
  write(txt);
  textattr:=oldcolor;
end;
{ ===================== cwriteln ================================== }
procedure cwriteln(hue : byte; txt : string);
 {set color & writeln }
var oldcolor : byte;
begin
  oldcolor:=textattr;
  textcolor(hue);
  writeln(txt);
  textattr:=oldcolor;
end;
{ ===================== put_text ================================== }
procedure put_text(col,row : integer; txt : string);
begin
   gotoxy(col,row);
   write(txt);
end;
{ ===================== put_line ================================== }
procedure put_line(col,row : integer; txt : string);
begin
   put_text(col,row,txt);
   writeln;
end;

{====================== windo.init ============================= }
constructor windo.init(tc,tr,bc,br,style,tclr,bclr : byte);
var
  m,n,h,i : word;
  j,x,y   : word;
  wrd : word;
  vmem : ^byte;
  hptr : ^byte;
  shadow_it : boolean;
begin;
  wstyle:=style;
  shadow_it:=true;
  if wstyle > 19 then wstyle:=10;
  {an illegal style yields no border & no shadow}
  if wstyle > 9 then
  begin
    shadow_it:=false;
    dec(wstyle,10);
  end;
  if br-1>hi(windmax) then br:=hi(windmax)+1;
  if bc-1>lo(windmax) then bc:=lo(windmax)+1;
  x := 1;
  {following reflects +1 on width & +1 on height  for shadow    }
  vbuffsize:=(2*((2+bc-tc)*(2+br-tr)));
  {it gets saved & restored whether a shadow is used or not ....}
  m := br - tr + 2;   { # of rows}
  n := bc - tc + 2;   { # of columns }
  vmem_offset(tc,tr,j);   { beginning location in vid memory }
  vmem := ptr(vidseg,j);
  h := 0;
  if (maxavail>vbuffsize) then
  getmem(vbuff,vbuffsize) else vbuffsize:=0;
  {if we can't save it all .... we won't save any}
  oldWmin:=WindMin;
  oldWmax:=WindMax;
  oldx := wherex;
  oldy := wherey;
  oldtattr:=textattr;
  topc:=tc;
  topr:=tr;
  botc:=bc;
  botr:=br;
  hptr:=vbuff;
  if vbuffsize>0 then
  repeat
    inc(h);
    move(vmem^,hptr^,n*2);
    inc(vmem,(160));
    inc(hptr,(n*2));
    inc(x,(n*2));
  until h = m;
  textcolor(tclr);
  textbackground(bclr);
  window(tc,tr,bc,br);
  wtxtattr:=textattr;
  clrscr;
  if wstyle > 0 then
     begin
        wind_brdr;
        window(tc+ 2,tr+1,bc- 2,br - 1);
     end
  else
  if wstyle=0 then window(tc,tr,bc,br);
  if shadow_it then
  begin
    shadow(tc+1,br+1,1+(bc-tc));
    for y:=(tr+1) to (br+1) do shadow(bc+1,y,1);
  end;
  wmin:=windmin;
  wmax:=windmax;
end;


{==================== windo.outer ==================}
procedure windo.outer;
begin
  if (windmax=wmax) and
     (windmin=wmin) then
  begin
    winx:=wherex;
    winy:=wherey;
  end;
  textattr:=oldtattr;
  windMin:=oldWmin;
  windMax:=oldWmax;
end;

{==================== windo.inner ==================}
procedure windo.inner;
begin
  textattr:=wtxtattr;
  windmax:=wmax;
  windmin:=wmin;
  gotoxy(winx,winy);
end;
{=============================== windo.WIND_BRDR =============================}
procedure windo.wind_brdr;
var
  tc,bc : byte;
  j : word;
  hlngth,vlngth : integer;
  tlcorn,trcorn,blcorn,brcorn : integer;
  thorlin,bhorlin,vertlin : integer;

  procedure brdr_style(var hlngth,vlngth,tlcorn,trcorn,
                       blcorn,brcorn,thorlin,
                       bhorlin,vertlin : integer);
  begin
    case (wstyle) of
          0 :  begin  { no border characters}
                 tlcorn      := 32;
                 thorlin     := 32;
                 trcorn      := 32;
                 vertlin     := 32;
                 blcorn      := 32;
                 bhorlin     := 32;
                 brcorn      := 32;
               end;
          1 :  begin { double line border}
                tlcorn      := 201;
                thorlin     := 205;
                   trcorn      := 187;
                   vertlin     := 186;
                   blcorn      := 200;
                   bhorlin     := 205;
                   brcorn      := 188;
               end;

         2 :  begin { single line border}
                   tlcorn      := 218;
                   thorlin     := 196;
                   trcorn      := 191;
                   vertlin     := 179;
                   blcorn      := 192;
                   bhorlin     := 196;
                   brcorn      := 217;
               end;

         3 :  begin { single line side, double line top & bottom }
                   tlcorn      := 213;
                   thorlin     := 205;
                   trcorn      := 184;
                   vertlin     := 179;
                   blcorn      := 212;
                   bhorlin     := 205;
                   brcorn      := 190;
               end;

         4 :  begin { double line side, single line top & bottom }
                   tlcorn      := 214;
                   thorlin     := 196;
                   trcorn      := 183;
                   vertlin     := 186;
                   blcorn      := 211;
                   bhorlin     := 196;
                   brcorn      := 189;
               end;

         5 :  begin { double line top, single line sides & bottom }
                   tlcorn      := 213;
                   thorlin     := 205;
                   trcorn      := 184;
                   vertlin     := 179;
                   blcorn      := 192;
                   bhorlin     := 196;
                   brcorn      := 217;
               end;
         6 :  begin { typewriter border, * in corners, - hor., | vert}
                   tlcorn      := ord('*');
                   thorlin     := ord('-');
                   trcorn      := tlcorn;
                   vertlin     := ord('|');
                   blcorn      := tlcorn;
                   bhorlin     := thorlin;
                   brcorn      := tlcorn;
               end;
         7 :  begin { typewriter border, + in corners, - hor., | vert}
                   tlcorn      := ord('+');
                   thorlin     := ord('-');
                   trcorn      := tlcorn;
                   vertlin     := ord('|');
                   blcorn      := tlcorn;
                   bhorlin     := thorlin;
                   brcorn      := tlcorn;
               end;
         8 :  begin { typewriter border, * in corners, = hor., I vert}
                   tlcorn      := ord('*');
                   thorlin     := ord('=');
                   trcorn      := tlcorn;
                   vertlin     := ord('I');
                   blcorn      := tlcorn;
                   bhorlin     := thorlin;
                   brcorn      := tlcorn;
              end;
         9 :  begin { typewriter border, * in corners, = hor., I vert}
                   tlcorn      := 220;
                   thorlin     := 220;
                   trcorn      := 220;
                   vertlin     := 219;
                   blcorn      := 223;
                   bhorlin     := 223;
                   brcorn      := 223;
              end;
         end;
   end;

begin
     tc:=topc+1;
     bc:=botc-1;
     hlngth := bc-tc+1;
     vlngth := botr-topr+1;
     brdr_style(hlngth,vlngth,tlcorn,trcorn,
                blcorn,brcorn,thorlin,
                bhorlin,vertlin);
     vmem_offset(tc,topr,j); {calc & place top left corner}
     chr_out(j,tlcorn);
     brdr_line(tc+1,topr,hlngth - 2, thorlin, 1); {do top horiz.}
     vmem_offset(bc,topr,j);  {calc & place top right corner}
     chr_out(j,trcorn);
     brdr_line(tc,topr+1, vlngth - 2, vertlin, 0); {do left vert.}
     vmem_offset(tc,botr,j);  {calc & place bottom left corner}
     chr_out(j,blcorn);
     brdr_line(bc,topr+ 1, vlngth - 2, vertlin, 0); {do right vert.}
     vmem_offset(bc,botr,j);  {calc & place bottom right corner}
     chr_out(j,brcorn);
     brdr_line(tc+ 1,botr, hlngth - 2, bhorlin, 1); {do bottom horiz.}
end;
{==================== windo.SHADOW ==================}
{ set the text to shadowed (grey on black) for a portion of a line}
procedure windo.SHADOW(col,row,span : byte);
var
  k   : word;
  i,j : word;
  b,f : byte;
  begin
    if col>lo(oldWmax)+1 then exit;
    if row>hi(oldWmax) then exit;
    while (col+span-2) > lo(oldwmax) do dec(span);
    vmem_offset(col,row,i);
    wrd_in(i,k);
    f:= 8;
    b := 0 and 240;{upper half is background ... 4 bits }
    for j := 1 to span do
     chr_out(i+(2*j)-1,f+b); {plug mod back into memory .....}
  end;
{====================== windo.done ============================= }
destructor windo.done;
var x,m,n : integer;
    j     : word;
    vmem  : ^byte;
    hptr  : ^byte;
    h     : integer;
begin
  x := 1;
  m := botr - topr + 2;   { # of rows}
  n := botc - topc + 2;   { # of columns }
  vmem_offset(topc,topr,j);   { beginning location in vid memory }
  vmem := ptr(vidseg,j);
  hptr := vbuff;
  h := 0;
  if vbuffsize>0 then
  begin
    repeat
      inc(h);
      move(hptr^,vmem^,n*2);
      inc(hptr,n*2);
      inc(vmem,160);
      inc(x,n*2);
    until h = m;
    freemem(vbuff,vbuffsize);
  end;
  WindMin := oldWmin;
  WindMax := oldWmax;
  textattr:= oldtattr;
  gotoxy(oldx,oldy-1);
end;
{====================== wait_for_a_key ============================= }
procedure wait_for_a_key(var keyread : char);
begin
  repeat
  until KeyPressed;
  keyread := readkey;
  if keyread = #0 then keyread := readkey;
end;
{======================= wait_for_any_key ============================}
procedure wait_for_any_key;
var
kr : char;
begin
  wait_for_a_key(kr);
end;

{------------------------------------------------------------------------}
begin
{$ifdef DPMI}
  vidseg:=SegB800;
{$else}
  vidseg:=$B800;
{$endif}
end.